import logging
import os
import pathlib
import shlex
import subprocess
import sys
from collections import OrderedDict
from configparser import _UNSET  # type: ignore
from configparser import ConfigParser
from configparser import NoOptionError
from configparser import NoSectionError
from typing import Dict
from typing import Optional
from typing import Union

from thabala_cli.exceptions import ThabalaCliConfigException
from thabala_cli.utils.module_loading import import_string

log = logging.getLogger(__name__)


def expand_env_var(env_var):
    """
    Expands (potentially nested) env vars by repeatedly applying
    `expandvars` and `expanduser` until interpolation stops having
    any effect.
    """
    if not env_var:
        return env_var
    while True:
        interpolated = os.path.expanduser(os.path.expandvars(str(env_var)))
        if interpolated == env_var:
            return interpolated
        env_var = interpolated


def run_command(command):
    """Runs command and returns stdout"""
    with subprocess.Popen(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
    ) as process:
        output, stderr = [
            stream.decode(sys.getdefaultencoding(), "ignore")
            for stream in process.communicate()
        ]
        if process.returncode != 0:
            raise ThabalaCliConfigException(
                f"Cannot execute {command}. Error code is: {process.returncode}. "
                f"Output: {output}, Stderr: {stderr}"
            )

    return output


def _default_config_file_path(file_name: str):
    templates_dir = os.path.join(os.path.dirname(__file__), "config_templates")
    return os.path.join(templates_dir, file_name)


# pylint: disable=too-many-ancestors
class ThabalaCliConfigParser(ConfigParser):
    """Custom Thabala CLI Configparser supporting defaults and deprecated options"""

    # These configuration elements can be fetched as the stdout of commands
    # following the "{section}__{name}__cmd" pattern, the idea behind this
    # is to not store password on boxes in text files.
    sensitive_config_values = set()  # type: ignore

    # This method transforms option names on every read, get, or set operation.
    # This changes from the default behaviour of ConfigParser from lowercasing
    # to instead be case-preserving
    def optionxform(self, optionstr: str) -> str:
        return optionstr

    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, default_config=None, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.thabala_cli_defaults = ConfigParser(*args, **kwargs)
        if default_config is not None:
            self.thabala_cli_defaults.read_string(default_config)

        self.is_validated = False

    def validate(self):
        self._validate_config_dependencies()
        self.is_validated = True

    def _validate_config_dependencies(self):
        """
        Validate that config values aren't invalid given other config values
        or system-level limitations and requirements.
        """
        # pylint: disable=unnecessary-pass
        pass

    @staticmethod
    def _env_var_name(section, key):
        return f"THABALA__{section.upper()}__{key.upper()}"

    def _get_env_var_option(self, section, key):
        # must have format THABALA__{SECTION}__{KEY} (note double underscore)
        env_var = self._env_var_name(section, key)
        if env_var in os.environ:
            return expand_env_var(os.environ[env_var])
        # alternatively THABALA__{SECTION}__{KEY}_CMD (for a command)
        env_var_cmd = env_var + "_CMD"
        if env_var_cmd in os.environ:
            # if this is a valid command key...
            if (section, key) in self.sensitive_config_values:
                return run_command(os.environ[env_var_cmd])
        return None

    def _get_cmd_option(self, section, key):
        fallback_key = key + "_cmd"
        # if this is a valid command key...
        if (section, key) in self.sensitive_config_values:
            if super().has_option(section, fallback_key):
                command = super().get(section, fallback_key)
                return run_command(command)
        return None

    def get(self, section, key, **kwargs):
        section = str(section).lower()
        key = str(key).lower()

        option = self._get_environment_variables(key, section)
        if option is not None:
            return option

        option = self._get_option_from_config_file(key, kwargs, section)
        if option is not None:
            return option

        option = self._get_option_from_commands(key, section)
        if option is not None:
            return option

        return self._get_option_from_default_config(section, key, **kwargs)

    def _get_option_from_default_config(self, section, key, **kwargs):
        # ...then the default config
        if self.thabala_cli_defaults.has_option(section, key) or "fallback" in kwargs:
            return expand_env_var(self.thabala_cli_defaults.get(section, key, **kwargs))

        log.warning("section/key [%s/%s] not found in config", section, key)

        raise ThabalaCliConfigException(
            f"section/key [{section}/{key}] not found in config"
        )

    def _get_option_from_commands(self, key, section):
        # ...then commands
        option = self._get_cmd_option(section, key)
        if option:
            return option
        return None

    def _get_option_from_config_file(self, key, kwargs, section):
        # ...then the config file
        if super().has_option(section, key):
            # Use the parent's methods to get the actual config here to be able to
            # separate the config from default config.
            return expand_env_var(super().get(section, key, **kwargs))
        return None

    def _get_environment_variables(self, key, section):
        # first check environment variables
        option = self._get_env_var_option(section, key)
        if option is not None:
            return option
        return None

    def getboolean(self, section, key, **kwargs):
        val = str(self.get(section, key, **kwargs)).lower().strip()
        if "#" in val:
            val = val.split("#")[0].strip()
        if val in ("t", "true", "1"):
            return True
        if val in ("f", "false", "0"):
            return False
        raise ThabalaCliConfigException(
            f'Failed to convert value to bool. Please check "{key}" key in "{section}" section. '
            f'Current value: "{val}".'
        )

    # pylint: disable=no-self-use
    def getint(self, section, key, **kwargs):
        val = self.get(section, key, **kwargs)

        try:
            return int(val)
        except ValueError as value_err:
            raise ThabalaCliConfigException(
                f'Failed to convert value to int. Please check "{key}" key in "{section}" section. '
                f'Current value: "{val}".'
            ) from value_err

    def getfloat(self, section, key, **kwargs):
        val = self.get(section, key, **kwargs)

        try:
            return float(val)
        except ValueError as value_err:
            raise ThabalaCliConfigException(
                f'Failed to convert value to float. Please check "{key}" key in "{section}" section. '
                f'Current value: "{val}".'
            ) from value_err

    def getimport(self, section, key, **kwargs):  # noqa
        """
        Reads options, imports the full qualified name, and returns the object.
        In case of failure, it throws an exception a clear message with the key aad the section names
        :return: The object or None, if the option is empty
        """
        full_qualified_path = conf.get(section=section, key=key, **kwargs)
        if not full_qualified_path:
            return None

        try:
            return import_string(full_qualified_path)
        except ImportError as import_err:
            log.error(import_err)
            raise ThabalaCliConfigException(
                f'The object could not be loaded. Please check "{key}" key in "{section}" section. '
                f'Current value: "{full_qualified_path}".'
            ) from import_err

    def read(self, filenames, encoding=None):
        super().read(filenames=filenames, encoding=encoding)

    def read_dict(self, dictionary, source="<dict>"):
        super().read_dict(dictionary=dictionary, source=source)

    def has_option(self, section, option):
        try:
            # Using self.get() to avoid reimplementing the priority order
            # of config variables (env, config, cmd, defaults)
            # UNSET to avoid logging a warning about missing values
            self.get(section, option, fallback=_UNSET)
            return True
        except (NoOptionError, NoSectionError):
            return False

    def remove_option(self, section, option, remove_default=True):
        """
        Remove an option if it exists in config from a file or
        default config. If both of config have the same option, this removes
        the option in both configs unless remove_default=False.
        """
        if super().has_option(section, option):
            super().remove_option(section, option)

        if self.thabala_cli_defaults.has_option(section, option) and remove_default:
            self.thabala_cli_defaults.remove_option(section, option)

    def getsection(
        self, section: str
    ) -> Optional[Dict[str, Union[str, int, float, bool]]]:
        """
        Returns the section as a dict. Values are converted to int, float, bool
        as required.
        :param section: section from the config
        :rtype: dict
        """
        if not self.has_section(section) and not self.thabala_cli_defaults.has_section(
            section
        ):
            return None

        if self.thabala_cli_defaults.has_section(section):
            _section = OrderedDict(self.thabala_cli_defaults.items(section))
        else:
            _section = OrderedDict()

        if self.has_section(section):
            _section.update(OrderedDict(self.items(section)))

        section_prefix = f"THABALA__{section.upper()}__"
        for env_var in sorted(os.environ.keys()):
            if env_var.startswith(section_prefix):
                key = env_var.replace(section_prefix, "")
                if key.endswith("_CMD"):
                    key = key[:-4]
                key = key.lower()
                _section[key] = self._get_env_var_option(section, key)

        for key, val in _section.items():
            try:
                val = int(val)
            except ValueError:
                try:
                    val = float(val)
                except ValueError:
                    if val.lower() in ("t", "true"):
                        val = True
                    elif val.lower() in ("f", "false"):
                        val = False
            _section[key] = val
        return _section

    def as_dict(
        self,
        display_source=False,
        display_sensitive=False,
        raw=False,
        include_env=True,
        include_cmds=True,
    ) -> Dict[str, Dict[str, str]]:
        """
        Returns the current configuration as an OrderedDict of OrderedDicts.
        :param display_source: If False, the option value is returned. If True,
            a tuple of (option_value, source) is returned. Source is either
            'cli.cfg', 'default', 'env var', or 'cmd'.
        :type display_source: bool
        :param display_sensitive: If True, the values of options set by env
            vars and bash commands will be displayed. If False, those options
            are shown as '< hidden >'
        :type display_sensitive: bool
        :param raw: Should the values be output as interpolated values, or the
            "raw" form that can be fed back in to ConfigParser
        :type raw: bool
        :param include_env: Should the value of configuration from AIRFLOW__
            environment variables be included or not
        :type include_env: bool
        :param include_cmds: Should the result of calling any *_cmd config be
            set (True, default), or should the _cmd options be left as the
            command to run (False)
        :type include_cmds: bool
        :param include_secret: Should the result of calling any *_secret config be
            set (True, default), or should the _secret options be left as the
            path to get the secret from (False)
        :return: Dictionary, where the key is the name of the section and the content is
            the dictionary with the name of the parameter and its value.
        """
        config_sources: Dict[str, Dict[str, str]] = {}
        configs = [
            ("default", self.thabala_cli_defaults),
            ("cli.cfg", self),
        ]

        self._replace_config_with_display_sources(
            config_sources, configs, display_source, raw
        )

        # add env vars and overwrite because they have priority
        if include_env:
            self._include_envs(config_sources, display_sensitive, display_source, raw)

        # add bash commands
        if include_cmds:
            self._include_commands(
                config_sources, display_sensitive, display_source, raw
            )
        return config_sources

    def _include_commands(self, config_sources, display_sensitive, display_source, raw):
        for (section, key) in self.sensitive_config_values:
            opt = self._get_cmd_option(section, key)
            if not opt:
                continue
            if not display_sensitive:
                opt = "< hidden >"
            if display_source:
                opt = (opt, "cmd")
            elif raw:
                opt = opt.replace("%", "%%")
            config_sources.setdefault(section, OrderedDict()).update({key: opt})
            del config_sources[section][key + "_cmd"]

    def _include_envs(self, config_sources, display_sensitive, display_source, raw):
        for env_var in [
            os_environment
            for os_environment in os.environ
            if os_environment.startswith("THABALA__")
        ]:
            try:
                _, section, key = env_var.split("__", 2)
                opt = self._get_env_var_option(section, key)
            except ValueError:
                continue
            if not display_sensitive and env_var != "THABALA__CORE__UNIT_TEST_MODE":
                opt = "< hidden >"
            elif raw:
                opt = opt.replace("%", "%%")
            if display_source:
                opt = (opt, "env var")

            section = section.lower()
            # if we lower key for kubernetes_environment_variables section,
            # then we won't be able to set any Airflow environment
            # variables. Thabala only parse environment variables starts
            # with THABALA__. Therefore, we need to make it a special case.
            if section != "kubernetes_environment_variables":
                key = key.lower()
            config_sources.setdefault(section, OrderedDict()).update({key: opt})

    @staticmethod
    def _replace_config_with_display_sources(
        config_sources, configs, display_source, raw
    ):
        for (source_name, config) in configs:
            for section in config.sections():
                ThabalaCliConfigParser._replace_section_config_with_display_sources(
                    config, config_sources, display_source, raw, section, source_name
                )

    @staticmethod
    def _replace_section_config_with_display_sources(
        config, config_sources, display_source, raw, section, source_name
    ):
        sect = config_sources.setdefault(section, OrderedDict())
        for (k, val) in config.items(section=section, raw=raw):
            if display_source:
                val = (val, source_name)
            sect[k] = val

    def load_test_config(self):
        """
        Load the unit test configuration.
        Note: this is not reversible.
        """
        # remove all sections, falling back to defaults
        for section in self.sections():
            self.remove_section(section)

        # then read test config

        path = _default_config_file_path("default_cli_test.cfg")
        log.info("Reading default test configuration from %s", path)
        self.read_string(_parameterized_config_from_template("default_cli_test.cfg"))
        # then read any "custom" test settings
        log.info("Reading test configuration from %s", TEST_CONFIG_FILE)
        self.read(TEST_CONFIG_FILE)


def get_thabala_home():
    """Get path to Thabala Home"""
    return expand_env_var(os.environ.get("THABALA_HOME", "~/.config/thabala"))


def get_thabala_cli_config(thabala_home):
    """Get Path to cli.cfg path"""
    if "THABALA_CONFIG" not in os.environ:
        return os.path.join(thabala_home, "cli.cfg")
    return expand_env_var(os.environ["THABALA_CONFIG"])


def _parameterized_config_from_template(filename) -> str:
    TEMPLATE_START = (
        "# ----------------------- TEMPLATE BEGINS HERE -----------------------\n"
    )

    path = _default_config_file_path(filename)
    with open(path) as fh:
        for line in fh:
            if line != TEMPLATE_START:
                continue
            return parameterized_config(fh.read().strip())
    raise RuntimeError(f"Template marker not found in {path!r}")


def parameterized_config(template):
    """
    Generates a configuration from the provided template + variables defined in
    current scope
    :param template: a config content templated with {{variables}}
    """
    all_vars = {k: v for d in [globals(), locals()] for k, v in d.items()}
    return template.format(**all_vars)  # noqa


def get_thabala_cli_test_config(thabala_home):
    """Get path to cli_test.cfg"""
    if "THABALA_CLI_TEST_CONFIG" not in os.environ:
        return os.path.join(thabala_home, "cli_test.cfg")
    return expand_env_var(os.environ["THABALA_CLI_TEST_CONFIG"])


def initialize_config():
    """
    Load the config files
    Called for you automatically as part of the boot process.
    """
    global THABALA_HOME

    default_config = _parameterized_config_from_template("default_thabala_cli.cfg")

    conf = ThabalaCliConfigParser(default_config=default_config)

    if conf.getboolean("core", "unit_test_mode"):
        # Load test config only
        if not os.path.isfile(TEST_CONFIG_FILE):
            log.info("Creating new config file for unit tests in: %s", TEST_CONFIG_FILE)
            pathlib.Path(THABALA_HOME).mkdir(parents=True, exist_ok=True)

            with open(TEST_CONFIG_FILE, "w") as file:
                cfg = _parameterized_config_from_template("default_cli_test.cfg")
                file.write(cfg)

        conf.load_test_config()
    else:
        # Load normal config
        if not os.path.isfile(THABALA_CLI_CONFIG):
            log.info("Creating new config file in: %s", THABALA_CLI_CONFIG)
            pathlib.Path(THABALA_HOME).mkdir(parents=True, exist_ok=True)

            with open(THABALA_CLI_CONFIG, "w") as file:
                file.write(default_config)

        log.info("Reading the config from %s", THABALA_CLI_CONFIG)

        conf.read(THABALA_CLI_CONFIG)

        # They _might_ have set unit_test_mode in the cli.cfg, we still
        # want to respect that and then load the unittests.cfg
        if conf.getboolean("core", "unit_test_mode"):
            conf.load_test_config()

    return conf


# Setting THABALA_HOME and THABALA_CLI_CONFIG from environment variables, using
# "~/thabala" and "THABALA_HOME/cli.cfg" respectively as defaults

THABALA_HOME = get_thabala_home()
THABALA_CLI_CONFIG = get_thabala_cli_config(THABALA_HOME)

TEST_CONFIG_FILE = get_thabala_cli_test_config(THABALA_HOME)

conf = initialize_config()
conf.validate()



from itertools import filterfalse
from itertools import tee
from typing import Callable
from typing import Dict
from typing import Iterable

from thabala_cli.exceptions import ThabalaCliApiException


def partition(pred: Callable, iterable: Iterable):
    """User a predicate to partition entries into false entries and true entries"""
    iter_1, iter_2 = tee(iterable)
    return filterfalse(pred, iter_1), filter(pred, iter_2)


def raise_thabala_cli_api_exc(err: Exception, response_json: Dict = None) -> None:
    detail = None

    if response_json:
        detail = response_json.get("detail")

    raise ThabalaCliApiException(f"{err} - {detail}" if detail else err)



import logging

import yaml

from thabala_cli.commands.utils.api_client import ApiClient
from thabala_cli.exceptions import ThabalaCliInfraCodeException

log = logging.getLogger(__name__)


VALID_KINDS = ["Authenticator", "Users", "ServiceInstance"]


class InfraCode:
    def __init__(self, profile):
        self.api_client = ApiClient(profile)

        InfraCode.kind_appliers = {
            "Authenticator": self._apply_authenticator,
            "Users": self._apply_users,
            "ServiceInstance": self._apply_service_instance,
        }

    def _apply_authenticator(self, resource):
        log.info(f"Applying authenticator...")
        return self.api_client.post_infra(resource)

    def _apply_users(self, resource):
        log.info(f"Applying users...")
        return self.api_client.post_infra(resource)

    def _apply_service_instance(self, resource):
        service_id = resource.get("instance", {}).get("service_id")
        service_instance_name = resource.get("instance", {}).get("name")
        log.info(
            f"Applying service instance... ({service_id}: {service_instance_name})"
        )
        return self.api_client.post_infra(resource)

    @staticmethod
    def _validate_resource(resource):
        kind = resource.get("kind")
        if not kind:
            raise ThabalaCliInfraCodeException(f"Resource kind not found in the YAML.")
        if kind not in InfraCode.kind_appliers.keys():
            raise ThabalaCliInfraCodeException(f"Invalid resource kind {kind}")

    @staticmethod
    def _load_yaml(yaml_file):
        resources = []
        try:
            with open(yaml_file, "r") as f:
                try:
                    resources_yaml = yaml.load_all(f, Loader=yaml.SafeLoader)
                    for resource_yaml in resources_yaml:
                        InfraCode._validate_resource(resource_yaml)
                        resources.append(resource_yaml)
                except yaml.YAMLError as exc:
                    raise ThabalaCliInfraCodeException(exc)
        except OSError as exc:
            raise ThabalaCliInfraCodeException(exc)

        return resources

    @staticmethod
    def apply_yaml(yaml_file):
        resources = InfraCode._load_yaml(yaml_file)
        result = []
        for resource in resources:
            kind = resource["kind"]
            result.append(InfraCode.kind_appliers[kind](resource))

        print(ApiClient.pretty_json(result))



"""Exceptions used by Thabala CLI"""


class ThabalaCliException(Exception):
    """
    Base class for all Thabala CLI errors.
    Each custom exception should be derived from this class
    """

    status_code = 500


class ThabalaCliConfigException(ThabalaCliException):
    """Raise when there is configuration problem"""


class ThabalaCliApiException(ThabalaCliException):
    """Raise when there is API response problem"""


class ThabalaCliInfraCodeException(ThabalaCliException):
    """Raise when there is an infra code related problem"""


class ThabalaOperationException(ThabalaCliException):
    """Raise when there is an operation problem"""


class ThabalaNotImplementedException(ThabalaCliException):
    """Raise when functionality is not implemented"""



"""Thabala logging settings"""
from typing import Any
from typing import Dict

from thabala_cli.configuration import conf
from thabala_cli.exceptions import ThabalaCliException

# TODO: Logging format and level should be configured
# in this file instead of from thabala.cfg. Currently
# there are other log format and level configurations in
# settings.py and cli.py.
LOG_LEVEL: str = conf.get("logging", "LOGGING_LEVEL").upper()

LOG_FORMAT: str = conf.get("logging", "LOG_FORMAT")

COLORED_LOG_FORMAT: str = conf.get("logging", "COLORED_LOG_FORMAT")

COLORED_LOG: bool = conf.getboolean("logging", "COLORED_CONSOLE_LOG")

COLORED_FORMATTER_CLASS: str = conf.get("logging", "COLORED_FORMATTER_CLASS")

BASE_LOG_FOLDER: str = conf.get("logging", "BASE_LOG_FOLDER")

DEFAULT_LOGGING_CONFIG: Dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "thabala_cli": {"format": LOG_FORMAT},
        "thabala_cli_colored": {
            "format": COLORED_LOG_FORMAT if COLORED_LOG else LOG_FORMAT,
            "class": COLORED_FORMATTER_CLASS if COLORED_LOG else "logging.Formatter",
        },
    },
    "handlers": {
        "console": {
            "class": "thabala_cli.utils.log.logging_mixin.RedirectStdHandler",
            "formatter": "thabala_cli_colored",
            "stream": "sys.stdout",
        },
    },
    "loggers": {},
    "root": {
        "handlers": ["console"],
        "level": LOG_LEVEL,
    },
}



"""Apply command"""
from thabala_cli.commands.utils.infra_code import InfraCode
from thabala_cli.commands.utils.profile import get_profile


def apply(args):
    """Apply infrastructure component"""
    profile = get_profile(args)
    infra_code = InfraCode(profile)

    infra_code.apply_yaml(args.file)



import logging

import requests

from thabala_cli.commands.utils.auth_base import AuthBase
from thabala_cli.utils.helpers import raise_thabala_cli_api_exc

log = logging.getLogger(__name__)


class AuthByPassword(AuthBase):
    """Authenticate by user and password via the default Thabala backend database"""

    def __init__(self, api_base_url, username, password):
        super().__init__(api_base_url)

        self._username = username
        self._password = password

    def authenticate(self):
        """User and password based authentication"""
        log.debug("Authenticating by password authenticator")

        try:
            response = requests.post(
                f"{self._api_base_url}/auth/token",
                headers=self._headers,
                json={
                    "grant_type": "password",
                    "username": self._username,
                    "password": self._password,
                },
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise_thabala_cli_api_exc(err, response.json())
        except Exception as err:
            raise_thabala_cli_api_exc(err)

        access_token = response.json()["access_token"]
        refresh_token = response.json()["refresh_token"]

        return access_token, refresh_token



import logging
from logging.config import dictConfig

from thabala_cli.configuration import conf
from thabala_cli.exceptions import ThabalaCliConfigException
from thabala_cli.utils.module_loading import import_string

log = logging.getLogger(__name__)


def configure_logging():
    """Configure & Validate Thabala logging"""
    logging_class_path = ""
    try:
        logging_class_path = conf.get("logging", "logging_config_class")
    except ThabalaCliConfigException:
        log.debug("Could not find key logging_config_class in config")

    if logging_class_path:
        try:
            logging_config = import_string(logging_class_path)

            # Make sure that the variable is in scope
            if not isinstance(logging_config, dict):
                raise ValueError("Logging Config should be of dict type")

            log.info(
                "Successfully imported user-defined logging config from %s",
                logging_class_path,
            )
        except Exception as err:
            # Import default logging configurations.
            raise ImportError(
                f"Unable to load custom logging from {logging_class_path} due to {err}"
            ) from err
    else:
        logging_class_path = "thabala_cli.config_templates.thabala_cli_logging_settings.DEFAULT_LOGGING_CONFIG"
        logging_config = import_string(logging_class_path)
        log.debug("Unable to load custom logging, using default config instead")

    try:
        # Try to init logging
        dictConfig(logging_config)
    except ValueError as value_err:
        log.error("Unable to load the config, contains a configuration error.")
        # When there is an error in the config, escalate the exception
        # otherwise Thabala would silently fall back on the default config
        raise value_err

    return logging_class_path



import urllib

from thabala_cli.exceptions import ThabalaNotImplementedException


class AuthBase:
    """Authenticator interface."""

    def __init__(self, api_base_url) -> None:
        self._api_base_url = api_base_url
        self._headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    def authenticate(self):
        raise ThabalaNotImplementedException



__all__ = ["version"]

try:
    from importlib import metadata  # type: ignore
except ImportError:  # for Python<3.8
    import importlib_metadata as metadata  # type: ignore

version = metadata.version("thabala-cli")
del metadata



import logging
import os
from typing import Optional

import pendulum

from thabala_cli.configuration import conf
from thabala_cli.logging_config import configure_logging

log = logging.getLogger(__name__)


TIMEZONE = pendulum.tz.timezone("UTC")
try:
    # pylint: disable=invalid-name
    tz = conf.getboolean("core", "default_timezone")
    if tz == "system":
        TIMEZONE = pendulum.tz.local_timezone()
    else:
        TIMEZONE = pendulum.tz.timezone(tz)
except Exception:  # pylint: disable=broad-except
    pass
log.info("Configured default timezone %s", TIMEZONE)

LOGGING_LEVEL = logging.INFO
LOG_FORMAT = conf.get("logging", "log_format")
SIMPLE_LOG_FORMAT = conf.get("logging", "simple_log_format")
LOGGING_CLASS_PATH: Optional[str] = None

USERNAME: Optional[str] = None
PASSWORD: Optional[str] = None


# pylint: disable=global-statement
def configure_vars():
    """Configure Global Variables from cli.cfg"""
    global USERNAME
    global PASSWORD
    USERNAME = conf.get("default", "USERNAME")
    PASSWORD = conf.get("default", "PASSWORD")


def initialize():
    """Initialize Thabala with all the settings from this file"""
    configure_vars()
    global LOGGING_CLASS_PATH
    LOGGING_CLASS_PATH = configure_logging()



import json
import logging
import os
import socket
import sys
import time
import urllib
import webbrowser

import requests

from thabala_cli.commands.utils.auth_base import AuthBase
from thabala_cli.exceptions import ThabalaCliApiException
from thabala_cli.exceptions import ThabalaOperationException
from thabala_cli.utils.helpers import raise_thabala_cli_api_exc

log = logging.getLogger(__name__)

BUF_SIZE = 16384


class AuthByWebBrowser(AuthBase):
    """Authenticates user by web browser."""

    def __init__(self, api_base_url, webbrowser_pkg=None, socket_pkg=None):
        super().__init__(api_base_url)

        self._protocol = None
        self._host = None
        self._port = None
        self._code = None
        self._state = None
        self._client_id = None
        self._consent_cache_id_token = True
        self._application = "Thabala CLI"
        self._webbrowser = webbrowser if webbrowser_pkg is None else webbrowser_pkg
        self._socket = socket.socket if socket_pkg is None else socket_pkg
        self._origin = None

    def authenticate(self):
        """Web browser based authentication"""
        log.debug("Authenticating by password authenticator")
        tokens = None

        socket_connection = self._socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            print(
                "Initiating login request with your identity provider. A "
                "browser window should have opened for you to complete the "
                "login. If you can't see it, check existing browser windows, "
                "or your OS settings. Press CTRL+C to abort and try again...",
                file=sys.stderr,
            )

            log.debug("step 1: query GS to obtain OAuth2 urls")
            authorize_url = self._get_authorize_url(callback_port=None)
            if not authorize_url:
                raise ThabalaOperationException(
                    "Authorized URL not found. Is your Thabala "
                    "instance configured to use external browser "
                    "based authentication?"
                )
            self._process_authorize_url(authorize_url)

            log.debug("step 2: start local server to handle callback url")
            try:
                socket_connection.bind(
                    (
                        os.getenv("THABALA_AUTH_SOCKET_ADDR", self._host),
                        int(os.getenv("THABALA_AUTH_SOCKET_PORT", self._port)),
                    )
                )
            except socket.gaierror as ex:
                if ex.args[0] == socket.EAI_NONAME:
                    raise ThabalaOperationException(
                        "localhost is not found. "
                        "Ensure /etc/hosts has localhost entry."
                    )
                else:
                    raise ex
            except TypeError as ex:
                if not self._host:
                    raise ThabalaCliApiException("Callback host is not defined")
                if not self._port:
                    raise ThabalaOperationException("Callback port is not defined")
                else:
                    raise ex
            socket_connection.listen(0)  # no backlog

            log.debug("step 3: open a browser")
            if not self._webbrowser.open_new(authorize_url):
                print(
                    "We were unable to open a browser window for you, "
                    "please open the following url manually then paste the "
                    "URL you are redirected to into the terminal.",
                    file=sys.stderr,
                )
                print(f"URL: {authorize_url}", file=sys.stderr)
                url = input("Enter the URL the authorize URL redirected you to: ")
                self._process_get_url(url)
                if not self._code or not self._state:
                    # Input contained no token, either URL was incorrectly pasted,
                    # empty or just wrong
                    raise ThabalaCliApiException(
                        "Unable to open a browser in this environment and "
                        "auth response contained no code or state."
                    )

            else:
                log.debug("step 4: accept token")
                self._receive_browser_code(socket_connection)
                tokens = self._get_tokens()

        finally:
            socket_connection.close()

        access_token = tokens["access_token"]
        refresh_token = tokens["refresh_token"]

        return access_token, refresh_token

    def _receive_browser_code(self, socket_connection):
        """Receives auth token from the web browser."""
        while True:
            socket_client, _ = socket_connection.accept()
            try:
                # Receive the data in small chunks and transmit it
                data = socket_client.recv(BUF_SIZE).decode("utf-8").split("\r\n")

                if not self._process_options(data, socket_client):
                    self._process_receive_browser_code(data, socket_client)
                    break
            finally:
                socket_client.shutdown(socket.SHUT_RDWR)
                socket_client.close()

    def _process_options(self, data, socket_client):
        """Allows JS Ajax access to this endpoint."""
        for line in data:
            if line.startswith("OPTIONS "):
                break
        else:
            return False

        AuthByWebBrowser._get_user_agent(data)
        requested_headers, requested_origin = AuthByWebBrowser._check_post_requested(
            data
        )
        if not requested_headers:
            return False

        if not self._validate_origin(requested_origin):
            # validate Origin and fail if not match with the server.
            return False

        self._origin = requested_origin
        content = [
            "HTTP/1.1 200 OK",
            "Date: {}".format(
                time.strftime("%a, %d %b %Y %H:%M:%S GMT", time.gmtime())
            ),
            "Access-Control-Allow-Methods: POST, GET",
            f"Access-Control-Allow-Headers: {requested_headers}",
            "Access-Control-Max-Age: 86400",
            f"Access-Control-Allow-Origin: {self._origin}",
            "",
            "",
        ]
        socket_client.sendall("\r\n".join(content).encode("utf-8"))
        return True

    def _validate_origin(self, requested_origin):
        ret = urllib.parse.urlsplit(requested_origin)
        netloc = ret.netloc.split(":")
        host_got = netloc[0]
        port_got = (
            netloc[1] if len(netloc) > 1 else (443 if self._protocol == "https" else 80)
        )

        return (
            ret.scheme == self._protocol
            and host_got == self._host
            and port_got == self._port
        )

    def _process_receive_browser_code(self, data, socket_client):
        if not self._process_get(data) and not self._process_post(data):
            return  # error

        content = [
            "HTTP/1.1 200 OK",
            "Content-Type: text/html",
        ]
        if self._origin:
            data = {"consent": self._consent_cache_id_token}
            msg = json.dumps(data)
            content.append(f"Access-Control-Allow-Origin: {self._origin}")
            content.append("Vary: Accept-Encoding, Origin")
        else:
            msg = """
            <!DOCTYPE html><html><head><meta charset="UTF-8"/>
            <title>Thabala Authentication Response</title></head>
            <body>
            Your identity was confirmed and propagated to {}.
            You can close this window now and go back where you started from.
            </body></html>""".format(
                self._application
            )
        content.append(f"Content-Length: {len(msg)}")
        content.append("")
        content.append(msg)

        socket_client.sendall("\r\n".join(content).encode("utf-8"))

    @staticmethod
    def _check_post_requested(data):
        request_line = None
        header_line = None
        origin_line = None
        for line in data:
            if line.startswith("Access-Control-Request-Method:"):
                request_line = line
            elif line.startswith("Access-Control-Request-Headers:"):
                header_line = line
            elif line.startswith("Origin:"):
                origin_line = line

        if (
            not request_line
            or not header_line
            or not origin_line
            or request_line.split(":")[1].strip() != "POST"
        ):
            return None, None

        return (
            header_line.split(":")[1].strip(),
            ":".join(origin_line.split(":")[1:]).strip(),
        )

    def _process_authorize_url(self, url: str) -> None:
        parsed_authorize_url = urllib.parse.parse_qs(urllib.parse.urlparse(url).query)
        if (
            "redirect_uri" not in parsed_authorize_url
            or "state" not in parsed_authorize_url
        ):
            return

        # Get callback URI properties
        parsed_redirect_uri = urllib.parse.urlparse(
            parsed_authorize_url["redirect_uri"][0]
        )
        if not parsed_redirect_uri.port:
            raise ThabalaOperationException(
                "Redirect URI callback port is not included in "
                f"the authentication URL: {url}"
            )
        self._protocol = parsed_redirect_uri.scheme
        self._host = parsed_redirect_uri.hostname
        self._port = parsed_redirect_uri.port

        # Get state
        self._state = parsed_authorize_url["state"][0]
        self._client_id = parsed_authorize_url["client_id"][0]

    def _process_get_url(self, url: str) -> None:
        parsed = urllib.parse.parse_qs(urllib.parse.urlparse(url).query)
        if "code" not in parsed or "state" not in parsed:
            return
        if not parsed["code"][0] or not parsed["state"][0]:
            return
        self._code = parsed["code"][0]

        if parsed["state"][0] != self._state:
            raise ThabalaCliApiException(
                "The requested authentication state " "doesn't match the received one"
            )

    def _process_get(self, data):
        for line in data:
            if line.startswith("GET "):
                target_line = line
                break
        else:
            return False

        self._get_user_agent(data)
        _, url, _ = target_line.split()
        self._process_get_url(url)
        return True

    def _process_post(self, data):
        for line in data:
            if line.startswith("POST "):
                break
        else:
            raise ThabalaCliApiException(
                "Invalid HTTP request from web browser. "
                "Idp authentication could have failed."
            )

        self._get_user_agent(data)
        try:
            # parse the response as JSON
            payload = json.loads(data[-1])
            parsed_state = payload.get("state")
            self._code = payload.get("code")
            self._consent_cache_id_token = payload.get("consent", True)
        except Exception:
            # key=value form.
            parsed_state = urllib.parse.parse_qs(data[-1])["state"][0]
            self._code = urllib.parse.parse_qs(data[-1])["code"][0]

        if parsed_state != self.state:
            raise ThabalaCliApiException(
                "Received state doesn't match the requested one"
            )

        return True

    @staticmethod
    def _get_user_agent(data):
        for line in data:
            if line.lower().startswith("user-agent"):
                log.debug(line)
                break
        else:
            log.debug("No User-Agent")

    def _get_authorize_url(self, callback_port=None):
        """Gets OAuth2 URLs from Thabala."""
        try:
            response = requests.post(
                f"{self._api_base_url}/auth/authenticator-request",
                headers=self._headers,
                json={"browser_mode_redirect_port": callback_port},
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise_thabala_cli_api_exc(err, response.json())
        except Exception as err:
            raise_thabala_cli_api_exc(err)

        oauth2 = response.json().get("oauth2", dict())
        authorize_url = oauth2.get("authorize_url")

        return authorize_url

    def _get_tokens(self):
        """Get thabala access and refresh tokens"""
        try:
            response = requests.post(
                f"{self._api_base_url}/auth/token",
                headers=self._headers,
                json={
                    "grant_type": "authorization_code",
                    "code": self._code,
                },
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise_thabala_cli_api_exc(err, response.json())
        except Exception as err:
            raise_thabala_cli_api_exc(err)

        return response.json()



import json
import logging
import os
import socket
import time
import urllib.parse
import webbrowser
from typing import Dict

import requests

from thabala_cli.commands.utils.auth_password import AuthByPassword
from thabala_cli.commands.utils.auth_webbrowser import AuthByWebBrowser
from thabala_cli.commands.utils.profile import EXTERNAL_BROWSER_AUTHENTICATOR
from thabala_cli.commands.utils.profile import PASSWORD_AUTHENTICATOR
from thabala_cli.exceptions import ThabalaCliApiException
from thabala_cli.exceptions import ThabalaOperationException
from thabala_cli.utils.helpers import raise_thabala_cli_api_exc

log = logging.getLogger(__name__)


POLL_INTERVAL_SECONDS = 3


class RequestState:
    """
    Static class with service instance action request constants
    """

    REQUESTED = "requested"
    QUEUED = "queued"
    INPROGRESS = "inprogress"
    COMPLETED = "completed"
    FAILED = "failed"

    service_actions = (REQUESTED, QUEUED, INPROGRESS, COMPLETED, FAILED)


class ApiClient:
    def __init__(self, profile):
        self.profile = profile
        self.api_base_url = f"{self.profile['account_url']}/api/v1"
        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

        self._access_token = None
        self._refresh_token = None

    @staticmethod
    def pretty_json(dic: Dict) -> str:
        return json.dumps(dic, indent=4, sort_keys=True)

    def _authenticate(self):
        """Login to the Thabala API"""
        authenticator = self.profile["authenticator"]

        if authenticator == PASSWORD_AUTHENTICATOR:
            auth_instance = AuthByPassword(
                self.api_base_url, self.profile["username"], self.profile["password"]
            )
        elif authenticator == EXTERNAL_BROWSER_AUTHENTICATOR:
            auth_instance = AuthByWebBrowser(self.api_base_url)
        else:
            raise ThabalaOperationException(f"Invalid authenticator: {authenticator}")

        self._access_token, self._refresh_token = auth_instance.authenticate()

    def _login_and_request(
        self, url: str, skip_login: bool = False, method: str = "GET", json_to_post=None
    ):
        if not skip_login:
            self._authenticate()

        try:
            response = requests.request(
                method,
                url,
                headers={
                    **self.headers,
                    **{
                        "Authorization": f"Bearer {self._access_token}",
                    },
                },
                json=json_to_post,
            )
            response.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise_thabala_cli_api_exc(err, response.json())
        except Exception as err:
            raise_thabala_cli_api_exc(err)

        return response.json()

    def _poll_service_instance_request(
        self,
        service_instance_id: str,
        service_instance_request_id: str,
        state_message: str = None,
    ):
        """Periodically poll the service instance requests API endpoint until
        the request got completed."""
        log.info(f"Polling request status... {state_message or ''}")

        response_json = self._login_and_request(
            f"{self.api_base_url}/service-instances/{service_instance_id}/requests/{service_instance_request_id}"
        )

        state = response_json.get("result", {}).get("state")
        if state == RequestState.FAILED:
            print(ApiClient.pretty_json(response_json))
            raise ThabalaCliApiException(f"Service instance request failed.")
        if state == RequestState.COMPLETED:
            return response_json

        time.sleep(POLL_INTERVAL_SECONDS)
        return self._poll_service_instance_request(
            service_instance_id, service_instance_request_id, state_message=f"({state})"
        )

    def get_users(self, limit: int, offset: int):
        """Get list of users by the Thabala API"""
        params = {
            "limit": f"limit={limit}" if limit else "",
            "offset": f"offset={offset}" if offset else "",
        }
        response_json = self._login_and_request(
            f"{self.api_base_url}/users?{params['limit']}&{params['offset']}"
        )
        print(ApiClient.pretty_json(response_json))

    def get_service_instances(
        self, limit: int, offset: int, service_id: str, service_instance_id: str
    ):
        """Get list of service instances by the Thabala API"""
        if service_instance_id:
            if limit or offset or service_id:
                raise_thabala_cli_api_exc(
                    Exception(
                        f"Limit, offset and service_id parameters are not "
                        f"allowed to use together with service_instance_id"
                    )
                )

            endpoint_url = f"service-instances/{service_instance_id}"
        else:
            params = {
                "limit": f"limit={limit}" if limit else "",
                "offset": f"offset={offset}" if offset else "",
                "service_id": f"service_id={service_id}" if service_id else "",
            }
            endpoint_url = f"service-instances?{params['limit']}&{params['offset']}&{params['service_id']}"

        response_json = self._login_and_request(f"{self.api_base_url}/{endpoint_url}")
        print(ApiClient.pretty_json(response_json))

    def get_service_instance_users(
        self, limit: int, offset: int, username, service_instance_id: str
    ):
        """Get users of service instances"""
        params = {
            "limit": f"limit={limit}" if limit else "",
            "offset": f"offset={offset}" if offset else "",
            "username": f"username={username}" if username else "",
            "service_instance_id": f"service_instance_id={service_instance_id}"
            if service_instance_id
            else "",
        }
        endpoint_url = f"service-instances/users?{params['limit']}&{params['offset']}&{params['username']}&{params['service_instance_id']}"

        response_json = self._login_and_request(f"{self.api_base_url}/{endpoint_url}")
        print(ApiClient.pretty_json(response_json))

    def pause_service_instance(self, service_instance_id: str):
        """pause a specific service instance"""
        log.info("Requesting service instance to pause...")
        response_json = self._login_and_request(
            f"{self.api_base_url}/service-instances/{service_instance_id}/pause",
            method="POST",
        )
        service_instance_request_id = response_json.get("service_instance_request_id")
        if not service_instance_request_id:
            raise ThabalaCliApiException(
                f"Invalid response. service_instance_request_id not received."
            )

        response_json = self._poll_service_instance_request(
            service_instance_id, service_instance_request_id
        )
        print(ApiClient.pretty_json(response_json))

    def resume_service_instance(self, service_instance_id: str):
        """Resume a specific service instance"""
        log.info("Requesting service instance to resume...")
        response_json = self._login_and_request(
            f"{self.api_base_url}/service-instances/{service_instance_id}/resume",
            method="POST",
        )
        service_instance_request_id = response_json.get("service_instance_request_id")
        if not service_instance_request_id:
            raise ThabalaCliApiException(
                f"Invalid response. service_instance_request_id not received."
            )

        response_json = self._poll_service_instance_request(
            service_instance_id, service_instance_request_id
        )
        print(ApiClient.pretty_json(response_json))

    def get_health(self):
        """Get health of Thabala account"""
        response_json = self._login_and_request(
            f"{self.api_base_url}/health", skip_login=True
        )
        print(ApiClient.pretty_json(response_json))

    def get_infra(self, kind):
        """Get infrastructure as a code of the Thabala account"""
        response_json = self._login_and_request(f"{self.api_base_url}/infra")
        kinds = []
        if kind:
            kinds = kind.split(",")

        for component in response_json.get("result", []):
            c_kind = component["kind"]
            code = component["code"]

            if not kinds or c_kind in kinds:
                print("---")
                print(code)

    def post_infra(self, resource):
        """Post an infrastructure resource"""
        response_json = self._login_and_request(
            f"{self.api_base_url}/infra", method="POST", json_to_post=resource
        )

        service_instance_id = response_json.get("service_instance", {}).get("id")
        service_instance_request_id = response_json.get("service_instance_request_id")
        if service_instance_id and service_instance_request_id:
            self._poll_service_instance_request(
                service_instance_id, service_instance_request_id
            )

        return response_json

    def get_version(self):
        """Get version of Thabala account"""
        response_json = self._login_and_request(
            f"{self.api_base_url}/version", skip_login=True
        )
        print(ApiClient.pretty_json(response_json))



"""Main executable module"""
import sys

import argcomplete

from thabala_cli import cli_parser
from thabala_cli.exceptions import ThabalaCliException


def main():
    """Main executable function"""
    try:
        parser = cli_parser.get_parser()
        argcomplete.autocomplete(parser)
        args = parser.parse_args()
        args.func(args)
    except ThabalaCliException as err:
        print(err)
        sys.exit(1)


if __name__ == "__main__":
    main()



"""Profile"""
from thabala_cli.configuration import conf
from thabala_cli.exceptions import ThabalaCliConfigException

PASSWORD_AUTHENTICATOR = "PASSWORD"
EXTERNAL_BROWSER_AUTHENTICATOR = "EXTERNALBROWSER"

AUTHENTICATORS = [
    PASSWORD_AUTHENTICATOR,
    EXTERNAL_BROWSER_AUTHENTICATOR,
]


def validate_profile(profile):
    if not profile.get("account_url"):
        raise ThabalaCliConfigException(f"Invalid profile. Missing account_url")
    if not profile.get("authenticator"):
        raise ThabalaCliConfigException(f"Invalid profile. Missing authenticator")

    #  Authenticator needs to be a known one
    if profile.get("authenticator") not in AUTHENTICATORS:
        raise ThabalaCliConfigException(
            f"Invalid authenticator: {profile.get('authenticator')}"
        )

    # Validate properties for the password authenticator
    if profile.get("authenticator") == PASSWORD_AUTHENTICATOR:
        if not profile.get("username"):
            raise ThabalaCliConfigException(f"Invalid profile. Missing username")
        if not profile.get("password"):
            raise ThabalaCliConfigException(f"Invalid profile. Missing password")


def get_profile(args):
    profile_name = (
        f"profile {args.profile}" if args.profile != "default" else args.profile
    )
    profile = {
        "account_url": conf.get(profile_name, "account_url"),
        "authenticator": conf.get(profile_name, "authenticator"),
        "username": conf.get(profile_name, "username"),
        "password": conf.get(profile_name, "password"),
    }

    validate_profile(profile)
    return profile



"""Service instance command"""
from thabala_cli.commands.utils.api_client import ApiClient
from thabala_cli.commands.utils.profile import get_profile


def pause(args):
    """Get users"""
    api_client = ApiClient(get_profile(args))
    api_client.pause_service_instance(args.service_instance_id)


def resume(args):
    """Get users"""
    api_client = ApiClient(get_profile(args))
    api_client.resume_service_instance(args.service_instance_id)



from setuptools import find_packages
from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements = fh.readlines()

setup(
    name="thabala-cli",
    version="1.0.1",
    python_requires=">= 3.7",
    description="Thabala Command Line Interface",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/koszti/thabala-cli",
    author="Thabala",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    packages=find_packages(exclude=["tests", "tests.*"]),
    include_package_data=True,
    zip_safe=False,
    install_requires=[req for req in requirements if req[:2] != "# "],
    entry_points={"console_scripts": ["thabala=thabala_cli.__main__:main"]},
)



"""Get command"""
from thabala_cli.commands.utils.api_client import ApiClient
from thabala_cli.commands.utils.profile import get_profile


def users(args):
    """Get users"""
    api_client = ApiClient(get_profile(args))
    api_client.get_users(args.limit, args.offset)


def service_instances(args):
    """Get service instances"""
    api_client = ApiClient(get_profile(args))
    api_client.get_service_instances(
        args.limit, args.offset, args.service_id, args.service_instance_id
    )


def service_instance_users(args):
    """Get service instance users"""
    api_client = ApiClient(get_profile(args))
    api_client.get_service_instance_users(
        args.limit, args.offset, args.username, args.service_instance_id
    )


def health(args):
    """Get account health"""
    api_client = ApiClient(get_profile(args))
    api_client.get_health()


def infra(args):
    """Get infrastructure as a code of the Thabala account"""
    api_client = ApiClient(get_profile(args))
    api_client.get_infra(args.kind)


def version(args):
    """Get account version"""
    api_client = ApiClient(get_profile(args))
    api_client.get_version()



"""Command-line interface"""
import argparse
from argparse import Action
from argparse import ArgumentParser
from argparse import RawTextHelpFormatter
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Union

from thabala_cli.utils.helpers import partition
from thabala_cli.utils.module_loading import import_string


def lazy_load_command(import_path: str) -> Callable:
    """Create a lazy loader for command"""
    _, _, name = import_path.rpartition(".")

    def command(*args, **kwargs):
        func = import_string(import_path)
        return func(*args, **kwargs)

    command.__name__ = name

    return command


# Used in Arg to enable `None' as a distinct value from "not passed"
_UNSET = object()


class Arg:
    """Class to keep information about command line argument"""

    # pylint: disable=redefined-builtin,unused-argument
    def __init__(
        self,
        flags=_UNSET,
        help=_UNSET,
        action=_UNSET,
        default=_UNSET,
        nargs=_UNSET,
        type=_UNSET,
        choices=_UNSET,
        required=_UNSET,
        metavar=_UNSET,
    ):
        self.flags = flags
        self.kwargs = {}
        for k, v in locals().items():
            if v is _UNSET:
                continue
            if k in ("self", "flags"):
                continue

            self.kwargs[k] = v

    def add_to_parser(self, parser: argparse.ArgumentParser):
        """Add this argument to an ArgumentParser"""
        parser.add_argument(*self.flags, **self.kwargs)


def positive_int(value):
    """Define a positive int type for an argument."""
    try:
        value = int(value)
        if value > 0:
            return value
    except ValueError:
        pass
    raise argparse.ArgumentTypeError(f"invalid positive int value: '{value}'")


# Shared
ARG_PROFILE = Arg(
    (
        "-p",
        "--profile",
    ),
    help="Thabala profile",
    default="default",
)
ARG_LIMIT = Arg(("--limit",), help="Number of items to return")
ARG_OFFSET = Arg(
    ("--offset",),
    help="Number of items to skip before starting to collect the result set",
)
ARG_SERVICE_ID = Arg(
    ("--service-id",),
    help="Service ID",
)
ARG_SERVICE_INSTANCE_ID = Arg(
    ("--service-instance-id",),
    help="Service instance ID",
)
ARG_SERVICE_INSTANCE_ID_REQUIRED = Arg(
    ("--service-instance-id",),
    help="Service instance ID",
    required=True,
)
ARG_USERNAME = Arg(
    ("-u", "--username"),
    help="Username",
)
ARG_KIND = Arg(
    ("--kind",),
    help="Infrastructure YAML kind. Allows multiple values by using comma separated list.",
)
ARG_FILE = Arg(
    ("-f", "--file"),
    help="Path to file",
    required=True,
)


# pylint: disable=inherit-non-class
class ActionCommand(NamedTuple):
    """Single CLI command"""

    name: str
    help: str
    func: Callable
    args: Iterable[Arg]
    description: Optional[str] = None
    epilog: Optional[str] = None


class GroupCommand(NamedTuple):
    """CLI command with subcommands"""

    name: str
    help: str
    subcommands: Iterable
    description: Optional[str] = None
    epilog: Optional[str] = None


CLICommand = Union[ActionCommand, GroupCommand]

GET_COMMANDS = (
    ActionCommand(
        name="infra",
        help="Get infrastructure as a code of the Thabala account",
        func=lazy_load_command("thabala_cli.commands.get_command.infra"),
        args=(
            ARG_PROFILE,
            ARG_KIND,
        ),
    ),
    ActionCommand(
        name="health",
        help="Get health af Thabala account",
        func=lazy_load_command("thabala_cli.commands.get_command.health"),
        args=(ARG_PROFILE,),
    ),
    ActionCommand(
        name="service-instances",
        help="Get Thabala service instances",
        func=lazy_load_command("thabala_cli.commands.get_command.service_instances"),
        args=(
            ARG_PROFILE,
            ARG_LIMIT,
            ARG_OFFSET,
            ARG_SERVICE_ID,
            ARG_SERVICE_INSTANCE_ID,
        ),
    ),
    ActionCommand(
        name="service-instance-users",
        help="Get users of service instances",
        func=lazy_load_command(
            "thabala_cli.commands.get_command.service_instance_users"
        ),
        args=(
            ARG_PROFILE,
            ARG_LIMIT,
            ARG_OFFSET,
            ARG_USERNAME,
            ARG_SERVICE_INSTANCE_ID,
        ),
    ),
    ActionCommand(
        name="users",
        help="Get Thabala users",
        func=lazy_load_command("thabala_cli.commands.get_command.users"),
        args=(
            ARG_PROFILE,
            ARG_LIMIT,
            ARG_OFFSET,
        ),
    ),
    ActionCommand(
        name="version",
        help="Get Thabala account version",
        func=lazy_load_command("thabala_cli.commands.get_command.version"),
        args=(ARG_PROFILE,),
    ),
)

SERVICE_INSTANCE_COMMANDS = (
    ActionCommand(
        name="pause",
        help="Pause a service instance",
        func=lazy_load_command("thabala_cli.commands.service_instance_command.pause"),
        args=(
            ARG_PROFILE,
            ARG_SERVICE_INSTANCE_ID_REQUIRED,
        ),
    ),
    ActionCommand(
        name="resume",
        help="Resume a service instance",
        func=lazy_load_command("thabala_cli.commands.service_instance_command.resume"),
        args=(
            ARG_PROFILE,
            ARG_SERVICE_INSTANCE_ID_REQUIRED,
        ),
    ),
)

thabala_cli_commands: List[CLICommand] = [
    GroupCommand(name="get", help="Get Thabala objects", subcommands=GET_COMMANDS),
    GroupCommand(
        name="service-instance",
        help="Service instance commands",
        subcommands=SERVICE_INSTANCE_COMMANDS,
    ),
    ActionCommand(
        name="apply",
        help="Apply infrastructure component",
        func=lazy_load_command("thabala_cli.commands.apply_command.apply"),
        args=(
            ARG_PROFILE,
            ARG_FILE,
        ),
    ),
    ActionCommand(
        name="version",
        help="Show the version",
        func=lazy_load_command("thabala_cli.commands.version_command.version"),
        args=(),
    ),
]
ALL_COMMANDS_DICT: Dict[str, CLICommand] = {sp.name: sp for sp in thabala_cli_commands}


class ThabalaHelpFormatter(argparse.HelpFormatter):
    """
    Custom help formatter to display help message.
    It displays simple commands and groups of commands in separate sections.
    """

    def _format_action(self, action: Action):
        if isinstance(
            action, argparse._SubParsersAction
        ):  # pylint: disable=protected-access

            parts = []
            action_header = self._format_action_invocation(action)
            action_header = "%*s%s\n" % (self._current_indent, "", action_header)
            parts.append(action_header)

            self._indent()
            subactions = action._get_subactions()  # pylint: disable=protected-access
            action_subcommands, group_subcommands = partition(
                lambda d: isinstance(ALL_COMMANDS_DICT[d.dest], GroupCommand),
                subactions,
            )
            parts.append("\n")
            parts.append("%*s%s:\n" % (self._current_indent, "", "Groups"))
            self._indent()
            for subaction in group_subcommands:
                parts.append(self._format_action(subaction))
            self._dedent()

            parts.append("\n")
            parts.append("%*s%s:\n" % (self._current_indent, "", "Commands"))
            self._indent()

            for subaction in action_subcommands:
                parts.append(self._format_action(subaction))
            self._dedent()
            self._dedent()

            # return a single string
            return self._join_parts(parts)

        return super()._format_action(action)


def get_parser() -> argparse.ArgumentParser:
    """Creates and returns command line argument parser"""
    parser = ArgumentParser(prog="thabala", formatter_class=ThabalaHelpFormatter)
    subparsers = parser.add_subparsers(dest="subcommand", metavar="GROUP_OR_COMMAND")
    subparsers.required = True

    subparser_list = ALL_COMMANDS_DICT.keys()
    sub_name: str
    for sub_name in sorted(subparser_list):
        sub: CLICommand = ALL_COMMANDS_DICT[sub_name]
        _add_command(subparsers, sub)
    return parser


def _sort_args(args: Iterable[Arg]) -> Iterable[Arg]:
    """Sort subcommand optional args, keep positional args"""

    def get_long_option(arg: Arg):
        """Get long option from Arg.flags"""
        return arg.flags[0] if len(arg.flags) == 1 else arg.flags[1]

    positional, optional = partition(lambda x: x.flags[0].startswith("-"), args)
    yield from positional
    yield from sorted(optional, key=lambda x: get_long_option(x).lower())


def _add_command(
    subparsers: argparse._SubParsersAction,
    sub: CLICommand,  # pylint: disable=protected-access
) -> None:
    sub_proc = subparsers.add_parser(
        sub.name,
        help=sub.help,
        description=sub.description or sub.help,
        epilog=sub.epilog,
    )
    sub_proc.formatter_class = RawTextHelpFormatter

    if isinstance(sub, GroupCommand):
        _add_group_command(sub, sub_proc)
    elif isinstance(sub, ActionCommand):
        _add_action_command(sub, sub_proc)
    else:
        raise Exception("Invalid command definition.")


def _add_action_command(sub: ActionCommand, sub_proc: argparse.ArgumentParser) -> None:
    for arg in _sort_args(sub.args):
        arg.add_to_parser(sub_proc)
    sub_proc.set_defaults(func=sub.func)


def _add_group_command(sub: GroupCommand, sub_proc: argparse.ArgumentParser) -> None:
    subcommands = sub.subcommands
    sub_subparsers = sub_proc.add_subparsers(dest="subcommand", metavar="COMMAND")
    sub_subparsers.required = True

    for command in sorted(subcommands, key=lambda x: x.name):
        _add_command(sub_subparsers, command)
